/*
% For use with Matlab
% Copyright (C) 2019 -- Jérôme Mendes <jermendes@gmail.com> or <jermendes@isr.uc.pt>, Rui Araújo <rui@isr.uc.pt>
% Jérôme Mendes, Francisco Souza, Rui Araújo, and Saeid Rastegar. "Neo-fuzzy neuron learning using backfitting algorithm". Neural Computing and Application, 2018.
% The "GAM-ZOTS toolbox" comes with ABSOLUTELY NO WARRANTY;
% In case of publication of any application of this method, please, cite the work:
% Jérôme Mendes, Francisco Souza, Rui Araújo, and Saeid Rastegar. 
% Neo-fuzzy neuron learning using backfitting algorithm.
% Neural Computing and Applications, 31(8):3609-3618, August 2019.
% DOI: https://doi.org/10.1007/s00521-017-3301-4
% http://www.isr.uc.pt/~jermendes/    ;     http://www.isr.uc.pt/~rui/
 **/

#include "math.h"
#include "mex.h"
#include <stdlib.h>
#include <iostream>
#include <vector>

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
//  MATLAB function x = AdditiveModels_C(data_ytrain, data_xtrain, x, limit_it, limit_error)
    
    /***** Inputs Declaration *****/
    const mxArray *data_ytrain, *data_xtrain, *x;
    double limit_it, limit_theta;

    data_ytrain = prhs[0];  // get input 1 (output data)
    data_xtrain = prhs[1];  // get input 2 (inputs data)
    x = prhs[2];            // get input 3 (struct with all parameters of the neo-fuzzy)
    limit_it = mxGetScalar(prhs[3]);    // get input 4 - variable lim_{it} 
    limit_theta = mxGetScalar(prhs[4]);  // get input 5 - variable \epsilon
    
    /****** Definition of the Output of this mex function*****/
    plhs[0] = mxDuplicateArray(x); // Duplicate the third function input (prhs[2]) (i.e. the actual struct with all parameters of the neo-fuzzy)

    /**** Local Variables Declaration *****/
    int n_data = mxGetM(data_xtrain);   //number (rows) of the data samples - variable K
    int n_var = mxGetN(data_xtrain);    //number (column) of the inputs variables - variable n
    int n_rules, j, i, k, it, count;// n_rules - number of rules, variable N_j; 
                                    // j is the input variable index, variable j; 
                                    // i is the rule index, variable i;
                                    // k is data index, variable k; 
                                    // it is variable it; 
                                    // count is an auxiliary variable
    mxArray *mx; // auxiliary variable
    double *Rules, *bias_j, *bias, *residual;   //Rules has dimensions (n_rules)x4: each row represents a rule; the first 3 columns are the 
                                            // antecedent membership function parameters (a_{j,i}, b_{j,i}, and c_{j,i} see Fig 2);
                                            // last column is the consequent parameter(\theta)
                                            // bias_j is the bias of the model of variable j
                                            // bias is the bias of the general model, variable y_0

    /****** Get matrix data *****/
    double *data_ytrain_Values = mxGetPr(data_ytrain); // get output data
    double *data_xtrain_Values = mxGetPr(data_xtrain); // get input data
    
    /****** Residual initialization *****/
    residual = (double *) mxCalloc(n_data, sizeof(double ));// variable y^l - Obtained on Algoritm 2 Step 3.c.A
    *residual = *data_ytrain_Values; // Initialized with the output data
    
    
    //////////////////////////////////////////////////////////////
    /////////////        Algoritm 2 Step 3.a.i       /////////////
    //////////////////////////////////////////////////////////////
    
    /**** Obtain bias of the general model ****/
    double mean = 0.0, mean_y = 0.0; //auxiliaries variables
    
    for (k=0;k<n_data;k++){ // for all output data
        
        mean_y = mean_y + data_ytrain_Values[k];
        
    }// end for (k=0;k<n_data;k++)
    
    mean_y = mean_y/n_data; // Obtain y_0 (Algoritm 2 Step 3.a.i)
    
    mx = mxGetField(plhs[0], 0, "bias");
    bias = mxGetPr(mx);
    bias[0] = mean_y; // save bias (variable y_0) on the struct (output)
    
    //////////////////////////////////////////////////////////////
    /////////////        Algoritm 2 Step 3.a.ii      /////////////
    //////////////////////////////////////////////////////////////
    
    /**** Initialization of thetas and bias_j by zero  ****/
    std::vector<double> theta_aux;  // auxiliary variable to obtain \theta (consequent parameters)
    count = 0;
    
    for (j=0;j<n_var;j++){// for all input variables
        
        mx = mxGetField(plhs[0], j, "Rules"); // Get the rules matrix of variable j
        Rules = mxGetPr(mx);
        
        n_rules = mxGetM(mx); // Obtain the number of rows (rules)
        
        // Set all thetas parameters to 0
        for (i=0;i<n_rules;i++){// for all rules of variable j
            
            Rules[n_rules*3 +i] = 0.0;
            theta_aux.push_back (0.0);
            count++;
        }
        // Set bias_j = 0.0
        mx = mxGetField(plhs[0], j, "bias_j"); // Get bias_j
        bias_j = mxGetPr(mx);
        bias_j[0] = 0.0;
        
    }// end for (j=0;j<n_var;j++)
    
    //////////////////////////////////////////////////////////////
    /////////////         Algoritm 2 Step 3.c        /////////////
    //////////////////////////////////////////////////////////////
    
    
    double Varvalue_j, aux, sum_mu, EstimationExcept_J[n_data], parameters[3], y_est_aux, y_est[n_data*(n_var-1)] ,epsilon;// auxiliaries variables
    int l, jj; // auxiliaries variables
    
    for (it=0;it<limit_it;it++){ // iterative learning (Algoritm 2 Step 3.c)
        
        for (j=0;j<n_var;j++){// for all input variables (Algoritm 2 Step 3.c.i)
            
            //////////////////////////////////////////////////////////////
            /////////////         Algoritm 2 Step 3.c.i.A    /////////////
            //////////////////////////////////////////////////////////////
            count = 0;
            
            for (jj=0;jj<n_var;jj++){// for all input variables
                
                if (j!=jj){ // Second term of equation presented on Algoritm 2 Step 3.c.i.A
                    
                    mx = mxGetField(plhs[0], jj, "Rules");  // Get the rules matrix of variable j
                    Rules = mxGetPr(mx);
                    n_rules = mxGetM(mx); // Obtain the number of rows (rules)
                    
                    double mu[n_rules];
                    
                    for (l=0;l<n_data;l++){// for all dataset
                        
                        Varvalue_j = data_xtrain_Values[n_data*jj + l]; // Obtain value of the input variable j on instant l of the dataset
                        
                        sum_mu = 0.0;
                        
                        for (i=0;i<n_rules;i++){// for all rules of variable j
                            parameters[0] = Rules[i];  // parameter a_{j,i};
                            parameters[1] = Rules[n_rules*1 +i];  // parameter b_{j,i};
                            parameters[2] = Rules[n_rules*2 +i];  // parameter c_{j,i};
                            
                            /**** Obtain the fuzzy triangular membership functions degree *****/
                            if (Varvalue_j > parameters[0] && Varvalue_j < parameters[1])
                                
                                mu[i] = (Varvalue_j-parameters[0])/(parameters[1]-parameters[0]);
                            
                            else if (Varvalue_j >= parameters[1] && Varvalue_j < parameters[2] )
                                
                                mu[i] = (parameters[2]-Varvalue_j)/ (parameters[2] - parameters[1] );
                            
                            else if (Varvalue_j <= parameters[0] || Varvalue_j >= parameters[2] )
                                
                                mu[i] = 0;
                            
                            /**** end ****/
                            
                            sum_mu = sum_mu +  mu[i];
                        }// end for (i=0;i<n_rules;i++)
                        y_est_aux = 0.0;

                        for (i=0;i<n_rules;i++){// for all rules of variable j
                            
                            epsilon = mu[i] / (sum_mu + 0.0000000000000000001); // variable \omega of the rule i of the variable j; last term is for protection of divide by zero
                            y_est_aux = y_est_aux + epsilon * Rules[n_rules*3 +i];
                            
                        }
                        
                        mx = mxGetField(plhs[0], jj, "bias_j");
                        bias_j = mxGetPr(mx);
                        
                        y_est[count] = y_est_aux - *bias_j; // obtain the estimated output of variable j
                        
                        count = count + 1;

                    }// end for (l=0;l<n_data;l++)
                    
                    
                }// end if (j!=jj)
                
            }// end for (jj=0;jj<n_var;jj++)

            for (l=0;l<n_data;l++){ // for all data
                
                aux = 0.0;
                
                for (count=0;count<n_var-1;count++){
                    
                    aux = aux + y_est[ l+count*n_data ];
                    
                }
                
                EstimationExcept_J[l] = aux; // second term of equation Algorithm 2 Step 3.c.i.A
                
                residual[l] = data_ytrain_Values[l] - mean_y - EstimationExcept_J[l]; // equation Algorithm 2 Step 3.c.i.A

            } //  end for (l=0;l<n_data;l++)
            
            //////////////////////////////////////////////////////////////
            /////////////         Algoritm 2 Step 3.c.i.B    /////////////
            //////////////////////////////////////////////////////////////
            
            
            mx = mxGetField(plhs[0], j, "Rules");// Get the rules matrix of variable j
            n_rules = mxGetM(mx);
            Rules = mxGetPr(mx);
            
            double mu[n_rules], epsilons_rule[n_data*n_rules];
            int ii;
            
            count = 0;
            
            for (ii=0;ii<n_rules;ii++){// for all rules of variable j

                for (l=0;l<n_data;l++){// for all dataset
                    
                    Varvalue_j = data_xtrain_Values[n_data*j + l];// Obtain value of the input variable j on instant l of the dataset
                    
                    sum_mu = 0.0;
                    
                    for (i=0;i<n_rules;i++){// for all rules of variable j
                        parameters[0] = Rules[i];// parameter a_{j,i};
                        parameters[1] = Rules[n_rules*1 +i];// parameter b_{j,i};
                        parameters[2] = Rules[n_rules*2 +i];// parameter c_{j,i};
                        
                        
                        /**** Obtain the fuzzy triangular membership functions degree ****/
                        if (Varvalue_j > parameters[0] && Varvalue_j < parameters[1])
                            
                            mu[i] = (Varvalue_j-parameters[0])/(parameters[1]-parameters[0]);
                        
                        else if (Varvalue_j >= parameters[1] && Varvalue_j < parameters[2] )
                            
                            mu[i] = (parameters[2]-Varvalue_j)/ (parameters[2] - parameters[1] );
                        
                        else if (Varvalue_j <= parameters[0] || Varvalue_j >= parameters[2] )
                            
                            mu[i] = 0;
                        
                        sum_mu = sum_mu +  mu[i];
                        /**** end ****/
                        
                    }
                    epsilons_rule[count] = mu[ii] / (sum_mu + 0.0000000000000000001); // variable \omega of the rule i of the variable j; last term is for protection of divide by zero
                    count++;
                    
                }// end for (l=0;l<n_data;l++)
            }// end for (ii=0;ii<n_rules;ii++)
            
            /**** To call regress function of the matlab  ****/
            mxArray  *thetareg;
            
            mxArray *lhs[2], *lhs_one, *lhs_zero;
            
            lhs_zero=mxCreateDoubleMatrix(n_data,1,mxREAL);
            lhs_one=mxCreateDoubleMatrix(n_data,n_rules,mxREAL);
            
            double *lhs_one_values = mxGetPr(lhs_one);
            double *lhs_zero_values = mxGetPr(lhs_zero);
            
            for (l=0;l<n_data;l++){
                lhs_zero_values[l] = residual[l];
                
            }

            count = 0;
            for (ii=0;ii<n_rules;ii++){
                
                for (l=0;l<n_data;l++){
                    lhs_one_values[count] = epsilons_rule[count];
                    count++;
                }
            }
            
            lhs[0] = lhs_zero;
            lhs[1] = lhs_one;
            
            mexCallMATLAB(1, &thetareg, 2, lhs, "regress");// regress Matlab function to obtain the consequent parameters (\theta), Algoritm 2 Step 3.c.i.B  
            
            double *thetareg_values=mxGetPr(thetareg);
            
            for (i=0;i<n_rules;i++){
                
                Rules[n_rules*3 +i] = thetareg_values[i];// save the new consequent parameters
            }
            
            //////////////////////////////////////////////////////////////
            /////////////         Algoritm 2 Step 3.c.i.C    /////////////
            //////////////////////////////////////////////////////////////

            double aux = 0.0;
            
            for (l=0;l<n_data;l++){// for all data
                
                Varvalue_j = data_xtrain_Values[n_data*j + l];// Obtain value of the input variable j on instant l of the dataset
                    
                sum_mu = 0.0;
                
                for (i=0;i<n_rules;i++){// for all rules of variable j
                    
                    parameters[0] = Rules[i];// parameter a_{j,i};
                    parameters[1] = Rules[n_rules*1 +i];// parameter b_{j,i};
                    parameters[2] = Rules[n_rules*2 +i];// parameter c_{j,i};
                    
                    /**** Obtain the fuzzy triangular membership functions degree ****/
                    if (Varvalue_j > parameters[0] && Varvalue_j < parameters[1])
                        
                        mu[i] = (Varvalue_j-parameters[0])/(parameters[1]-parameters[0]);
                    
                    else if (Varvalue_j >= parameters[1] && Varvalue_j < parameters[2] )
                        
                        mu[i] = (parameters[2]-Varvalue_j)/ (parameters[2] - parameters[1] );
                    
                    else if (Varvalue_j <= parameters[0] || Varvalue_j >= parameters[2] )
                        
                        mu[i] = 0;
                    
                    sum_mu = sum_mu +  mu[i];
                    /**** end ****/
                }
                
                y_est_aux = 0.0;
                
                for (i=0;i<n_rules;i++){
                    
                    epsilon = mu[i] / (sum_mu + 0.0000000000000000001);
                    
                    y_est_aux = y_est_aux + epsilon * Rules[n_rules*3 +i];
                }
                
                aux = aux + y_est_aux;

            }
            
            mean = aux/n_data; // bias of model of variable j
            
            // Set bias_j = mean 
            mx = mxGetField(plhs[0], j, "bias_j");
            bias_j = mxGetPr(mx);
            bias_j[0] = mean;
            
        }
        
        aux = 0.0;

        count=0;
        //////////////////////////////////////////////////////////////
        /////////////         Algoritm 2 Step 3.c.ii     /////////////
        //////////////////////////////////////////////////////////////
        
        for (j=0;j<n_var;j++){// for all input variables

            mx = mxGetField(plhs[0], j, "Rules");
            n_rules = mxGetM(mx);
            Rules = mxGetPr(mx);

            for (i=0;i<n_rules;i++){

                aux = aux + (  Rules[n_rules*3 +i] -theta_aux[count] ) * (  Rules[n_rules*3 +i] -theta_aux[count] ) ; // Cal the morm of gammas
                
                theta_aux[count] =  Rules[n_rules*3 +i] ;// Save the actual gamma
                
                count++;
            }
            
        }
        aux = sqrt(aux);
        
        if(aux < limit_theta){
            
            it = limit_it; // Termination condition
        }
    }
    
    return;
}
